/*
 *	chkey_common.c
 *
 *	Copyright (c) 1988-1992 Sun Microsystems Inc
 *	All Rights Reserved.
 */

#pragma ident "@(#)chkey_common.c	1.5	92/04/21 SMI"

#include <stdio.h>
#include <pwd.h>
#include <shadow.h>
#include <netdb.h>
#include <mp.h>
#include <rpcsvc/nis.h>
#include <rpc/key_prot.h>
#include <nsswitch.h>


extern char *crypt();
extern long random();
extern char *getpass();
extern char program_name[256];
static const char* CRED_TABLE = "cred.org_dir";

#define ROOTKEY_FILE "/etc/.rootkey"

#ifndef MAXHOSTNAMELEN
#define	MAXHOSTNAMELEN 256
#endif

#define PK_FILES   1
#define PK_YP      2
#define PK_NISPLUS 3



/* ************************ switch functions *************************** */

/*	NSW_NOTSUCCESS  NSW_NOTFOUND   NSW_UNAVAIL    NSW_TRYAGAIN */
#define	DEF_ACTION {__NSW_RETURN, __NSW_RETURN, __NSW_CONTINUE, __NSW_CONTINUE}

static struct __nsw_lookup lookup_files = {"files", DEF_ACTION, NULL, NULL},
		lookup_nis = {"nis", DEF_ACTION, NULL, &lookup_files};
static struct __nsw_switchconfig publickey_default =
			{0, "publickey", 2, &lookup_nis};


char*
switch_policy_str( struct __nsw_switchconfig *conf )
{
	struct __nsw_lookup *look;
	static char policy[256];
	int previous = 0;

	memset( (char*)policy, 0, 256 );

	for (look = conf->lookups; look; look = look->next) {
		if( previous )
			strcat( policy, " ");
		strcat( policy, look->service_name );
		previous = 1;
	}

	return( policy );
}

int
no_switch_policy( struct __nsw_switchconfig *conf )
{
	return( conf == NULL || conf->lookups == NULL );
}

is_switch_policy( struct __nsw_switchconfig *conf, char* target )
{
	return( conf && 
	       conf->lookups && 
	       strcmp(conf->lookups->service_name, target) == 0 &&
	       conf->lookups->next == NULL );
}

char*
first_and_only_switch_policy( char* policy, 
		    struct __nsw_switchconfig *default_conf,
		    char* head_msg )
{
	struct __nsw_switchconfig *conf;
	enum __nsw_parse_err perr;
	int policy_correct = 1;
	char* target_service = 0;
	int use_default = 0;

	if( default_conf == 0 )
		default_conf = &publickey_default;

	conf = __nsw_getconfig( policy , &perr);
	if( no_switch_policy( conf ) ) {
		use_default = 1;
		conf = default_conf;
	}

	target_service = conf->lookups->service_name;

	if( conf->lookups->next != NULL ) {
		policy_correct = 0;
		if( use_default ) {
			fprintf(stderr,
			"\n%s\n There is no publickey entry in %s.\n",
				head_msg, __NSW_CONFIG_FILE);
			fprintf(stderr,
			" The default publickey policy is \"publickey: %s\".\n",
			switch_policy_str( default_conf ));
		} else 
			fprintf(stderr,
		"\n%s\n The publickey entry in %s is \"publickey: %s\".\n",
			head_msg, __NSW_CONFIG_FILE,
			switch_policy_str( conf ));
	}

	if( policy_correct == 0 )
		fprintf(stderr, 
		" I cannot figure out which policy you want to update.\n");
        if( !use_default && conf )
		__nsw_freeconfig(conf);

	if( policy_correct )
		return( target_service );
	else
		return(0);
}



int
check_switch_policy( char* policy, char* target_service,
		    struct __nsw_switchconfig *default_conf,
		    char* head_msg, char* tail_msg )
{
	struct __nsw_switchconfig *conf;
	enum __nsw_parse_err perr;
	int policy_correct = 1;

	if( default_conf == 0 )
		default_conf = &publickey_default;

	conf = __nsw_getconfig( policy , &perr);
	if( no_switch_policy( conf ) ) {
		if( !is_switch_policy( default_conf, target_service )) {
			fprintf(stderr,
				"\n%s\n There is no publickey entry in %s.\n",
				head_msg, __NSW_CONFIG_FILE);
			fprintf(stderr,
			" The default publickey policy is \"publickey: %s\".\n",
			switch_policy_str( default_conf ));
			policy_correct = 0;
		}
	} else if (!is_switch_policy( conf, target_service )) {
		fprintf(stderr,
		"\n%s\n The publickey entry in %s is \"publickey: %s\".\n",
			head_msg, __NSW_CONFIG_FILE,
			switch_policy_str( conf ));
		policy_correct = 0;
	}
	/* should we exit ? */
	if( policy_correct == 0 )
		fprintf(stderr, 
		" It should be \"publickey: %s\"%s\n\n",
			target_service, tail_msg );
        if( conf )
		__nsw_freeconfig(conf);

	return( policy_correct );
}

int
get_pk_source( char* pk_service )
{
	int db = 0, got_from_switch = 0;

	/* No service specified, try to figure out from switch */
	if( pk_service == 0 ) {
		pk_service = first_and_only_switch_policy("publickey", 0,
							  "ERROR:");
		if( pk_service == 0 )
			return( 0 );
		fprintf(stderr, 
			"Updating %s publickey database.\n",
			pk_service);
		got_from_switch = 1;
	}

	if( strcmp( pk_service, "nisplus" ) == 0 )
		db = PK_NISPLUS;
	else if( strcmp( pk_service, "nis" ) == 0 )
		db = PK_YP;
	else if( strcmp( pk_service, "files" ) == 0 )
		db = PK_FILES;
	else return( 0 );

	/* If we didn't get service name from switch, check switch
	   and print warning about it source of publickeys if not unique */
	if( got_from_switch == 0 )
		check_switch_policy("publickey", pk_service, 0,"WARNING:", 
				    db == PK_FILES ? "" :
			    "; add 'files' if you want the 'nobody' key.");

	
	return( db ); /* all passed */
}


/* ***************************** keylogin stuff *************************** */
int
keylogin( char* netname, char* secret)
{
	struct key_netstarg netst;

	netst.st_pub_key[0] = 0;
	strcpy(netst.st_priv_key, secret);
	netst.st_netname = netname;

#ifdef NFS_AUTH
	nra.authtype = AUTH_DES;	/* only revoke DES creds */
	nra.uid = getuid();		/* use the real uid */
	if (_nfssys(NFS_REVAUTH, &nra) < 0) {
		perror("Warning: NFS credentials not destroyed");
		err = 1;
	}
#endif NFS_AUTH


	/* do actual key login */
	if (key_setnet(&netst) < 0 ) {
		fprintf(stderr, "Could not set %s's secret key\n", netname);
		fprintf(stderr, "May be the keyserv is down?\n");
		return( 0 );
	}

	return( 1 );
}

/* write unencrypted secret key into root key file */
write_rootkey( char* secret )
{
	char sbuf[HEXKEYBYTES+2];
	int fd, len;
	

	strcpy(sbuf, secret);
	strcat(sbuf, "\n");
	len = strlen(sbuf);
	sbuf[len] = '\0';
	unlink( ROOTKEY_FILE );
	if( (fd = open(ROOTKEY_FILE, O_WRONLY+O_CREAT, 0600)) != -1 ) {
		write(fd, sbuf, len+1);
		close(fd);
		fprintf(stderr, "Wrote secret key into %s\n",
			ROOTKEY_FILE);
	} else {
		fprintf(stderr, "Could not open %s for update\n",
			ROOTKEY_FILE );
		perror(ROOTKEY_FILE);
	}
}

char* 
get_password( uid_t uid, int same_host, char* target_host, int force )
{
	static char     password[256];
	char            prompt[256];
	char		*encrypted_password, *login_password = NULL, *pass;
	struct passwd	*pw;
	struct spwd     *spw;
	int passwords_matched = 0;

	uid_t my_uid;

	my_uid = geteuid();
	if( uid == 0 ) {
		if( same_host ) {
			pw = getpwuid(0);
			if (! pw) {
				fprintf(stderr, 
			"%s: unable to locate password record for uid %d\n", 
					program_name, uid);
				return (0);
			}
			spw = getspnam(pw->pw_name); 
			if( !spw ) {
				fprintf( stderr, 
			"%s: unable to locate password record for uid 0\n", 
					program_name);
				return (0);
			}
			login_password = spw->sp_pwdp;
		}
	} else {
		pw = getpwuid(uid);
		if (! pw) {
			fprintf(stderr, 
			"%s: unable to locate password record for uid %d\n", 
				program_name, uid);
			return (0);
		}
		if( pw->pw_passwd && strlen(pw->pw_passwd) != 0 )
			login_password = pw->pw_passwd;
		else if( my_uid == 0 ) {
			/* could not get password. If I'm root, try shadow */
			spw = getspnam(pw->pw_name);
			if( spw ) {
				login_password = spw->sp_pwdp;
			}
		}
	}


	if( (uid == my_uid) && ((uid != 0) || same_host ) )
		sprintf( prompt, "Enter login password:" );
	else if( uid == 0 ) {
		sprintf( prompt, "Enter %s's root login password:",
			target_host );
	} else
		sprintf( prompt, "Enter %s's login password:", 
			pw->pw_name );

	pass = getpass( prompt );
	if (strlen(pass) == 0) {
		(void)fprintf(stderr,"%s: Password unchanged.\n", 
			      program_name);
		return (0);
	}
	strcpy( password, pass );

	/* Verify that password supplied matches login password */
	if( login_password && (strlen(login_password) != 0)) {
		encrypted_password = crypt(password, login_password);
		if (strcmp(encrypted_password, login_password) == 0)
			passwords_matched = 1;
		else {
			fprintf(stderr, 
			"%s: %s: password differs from login password.\n",
				program_name, force? "WARNING" : "ERROR");
			if( !force )
				return(0);
		}
	}

	/* Check for mis-typed password */
	if( !passwords_matched ) {
		pass = getpass("Retype password:");
		if( strcmp( password, pass ) != 0 ) {
			(void)fprintf(stderr, "%s: password incorrect.\n", 
				      program_name);
			return (0);
		}
	}

	return( password );
}

/******************** netname generation *************************** */
#define MAXIPRINT 11
#define OPSYS_LEN 4
static const char* OPSYS = "unix";

getnetnameof( char* netname, uid_t uid, char* domain )
{
	int status;

	if (uid == 0) 
		status = host2netname(netname, (char *) NULL, domain );
	else {
		/* generate netname using uid and domain information. */
		int len;
		len = strlen(domain);
		if ((len + OPSYS_LEN + 3 + MAXIPRINT) > MAXNETNAMELEN) {
			printf( "Domain name too long: %s\n", domain );
			goto not_found;
		}
		(void) sprintf(netname, "%s.%d@%s", OPSYS, uid, domain);
		len = strlen(netname);
		if (netname[len-1] == '.')
			netname[len-1] = '\0';

		status = 1;
	}

	return (1);

not_found:
	return(-1);
}

/* ****************************** nisplus stuff ************************* */
/* most of it gotten from nisaddcred */


/* Check that someone else don't have the same auth information already */
static
nis_error
auth_exists( char* princname, char* auth_name, char* auth_type, char* domain)
{
	char sname[NIS_MAXNAMELEN+1];
	nis_result	*res;
	nis_error status;
	char* foundprinc;

	sprintf(sname,"[auth_name=%s,auth_type=%s],%s.%s",
		auth_name, auth_type, CRED_TABLE, domain );
	if (sname[strlen(sname)-1] != '.')
		strcat(sname, ".");
	/* Don't want FOLLOW_PATH here */
	res = nis_list(sname, 
		       MASTER_ONLY+USE_DGRAM+NO_AUTHINFO+FOLLOW_LINKS,
		       NULL, NULL);

	status = res->status;
	switch (res->status) {
	case NIS_NOTFOUND:
		break;
	case NIS_TRYAGAIN :
		fprintf(stderr, 
			"%s: NIS+ server busy, try again later.\n",
			program_name);
		exit(1);
	case NIS_PERMISSION :
		fprintf(stderr, 
			"%s: insufficent permission to look up old credentials.\n",
			program_name);
		exit(1);
	case NIS_SUCCESS:
		foundprinc = ENTRY_VAL(res->objects.objects_val, 0);
		if( nis_dir_cmp( foundprinc, princname ) != SAME_NAME ) {
			fprintf(stderr, 
    "%s: %s credentials with auth_name '%s' already belong to '%s'.\n",
				program_name, auth_type, auth_name, foundprinc);
			exit(1);
		}
		break;
	default:
		fprintf(stderr, 
			"%s: error looking at cred table, NIS+ error: %s\n",
			program_name, nis_sperrno(res->status));
		exit(1);
	}
	nis_freeresult( res );
	return( status );
}

static int
sanity_checks( char* nis_princ, char* netname, char* domain )
{
	char            netdomainaux[MAXHOSTNAMELEN+1];
	char            *princdomain, *netdomain;
	int             len;

	/* Sanity check 0. Do we have a nis+ principal name to work with? */
	if( nis_princ == NULL ) {
		fprintf(stderr, 
			"%s: you must create a \"local\" credential first.\n",
			program_name);
		fprintf(stderr,	"See nisaddcred(1).\n");
		return(0);
	}

	/* Sanity check 0.5.  NIS+ principal names must be dotted. */
	len = strlen(nis_princ);
	if( nis_princ[len-1] != '.' ) {
		fprintf(stderr,
		"%s: invalid principal name: '%s' (forgot ending dot?).\n",
			program_name, nis_princ);
		return (0);
	}

	/* Sanity check 1.  We only deal with one type of netnames. */
	if (strncmp(netname, OPSYS, OPSYS_LEN) != 0) {
		fprintf(stderr,"%s: unrecognized netname type: '%s'.\n",
			program_name, netname);
		return (0);
	}

	/* Sanity check 2.  Should only add DES cred in home domain. */
	princdomain = nis_domain_of( nis_princ );
	if( strcasecmp( princdomain, domain ) != 0 ) {
		fprintf(stderr,
"%s: domain of principal '%s' does not match destination domain '%s'.\n",
			nis_princ, domain );
		fprintf(stderr,
	"Should only add DES credential of principal in its home domain\n");
		return(0);
	}

	/* Sanity check 3:  Make sure netname's domain same as principal's
	   and don't have extraneous dot at the end.
	 */
	netdomain = (char*) strchr(netname, '@');
	if (! netdomain || netname[strlen(netname)-1] == '.' ) {
		fprintf(stderr, "%s: invalid netname: '%s'. \n", 
			program_name, netname );
		return (0);
	}
	netdomain++; /* skip '@' */
	strcpy( netdomainaux, netdomain );
	if( netdomainaux[strlen(netdomainaux)-1] != '.' )
		strcat( netdomainaux, "." );

	if( strcasecmp(princdomain,netdomainaux) != 0 ) {
		fprintf(stderr, 
	"%s: domain of netname %s should be same as that of principal %s\n",
			program_name, netname, nis_princ );
		return( 0 );
	}

	/* Another principal owns same credentials? (exits if that happens) */
	(void) auth_exists(nis_princ, netname, "DES", domain);

	return( 1 ); /* all passed */
}


/* Similar to nis_local_principal (nis_subr.c) except
 * this gets the results from the MASTER_ONLY and no FOLLOW_PATH.  
 * We only want the master because we'll be making updates there,
 * and also the replicas may not have seen the 'nisaddacred local'
 * that may have just occurred.
 * Returns NULL if not found.
 */
static
char*
default_principal( directory )
char* directory;
{
	nis_result	*res;
	char           buf[NIS_MAXNAMELEN+1];
	int            status;
	static char    principal_name[NIS_MAXNAMELEN+1];
	uid_t          uid;

	uid = geteuid();

	if( uid == 0 )
		return(nis_local_host());

	sprintf(buf, "[auth_name=%d,auth_type=LOCAL],%s.%s",
		uid, CRED_TABLE, directory );

	if (buf[strlen(buf)-1] != '.')
                strcat(buf, ".");

	res = nis_list(buf,
		       MASTER_ONLY+USE_DGRAM+NO_AUTHINFO+FOLLOW_LINKS,
		       NULL, NULL);
	
	if ( res == NULL ) {
		fprintf(stderr,
			"%s: unable to get result from NIS+ server.",
			program_name);
		exit(1);
	}
	switch( res->status ) {
	case NIS_SUCCESS:
		if(res->objects.objects_len > 1) {
			/* More than one principal with same uid?
			 * something wrong with cred table. Should be unique
			 * Warn user and continue. 
			 */
			fprintf(stderr, 
	       "%s: LOCAL entry for %d in directory %s not unique",
				program_name, uid, directory);
		}
		strcpy(principal_name,
		       ENTRY_VAL(res->objects.objects_val, 0));
		nis_freeresult(res);
		return( principal_name );

	case NIS_NOTFOUND:
		nis_freeresult(res);
		return( NULL );

	case NIS_TRYAGAIN :
		fprintf(stderr, 
			"%s: NIS+ server busy, try again later.\n",
			program_name);
		exit(1);

	case NIS_PERMISSION :
		fprintf(stderr, 
			"%s: insufficent permission to update credentials.\n",
			program_name);
		exit(1);

	default:
		fprintf(stderr, 
			"%s: error talking to server, NIS+ error: %s.\n",
			program_name, nis_sperrno(res->status));
		exit(1);
	}
	return(NULL);
}



/* Check whether this principal already has this type of credentials */
static nis_error
cred_exists( char* nisprinc, char* flavor, char* domain )
{
	char sname[NIS_MAXNAMELEN+1];
	nis_result	*res;
	nis_error status;

	sprintf(sname, "[cname=\"%s\",auth_type=%s],%s.%s", 
		nisprinc, flavor, CRED_TABLE, domain );
	if (sname[strlen(sname)-1] != '.')
                strcat(sname, ".");

	/* Don't want FOLLOW_PATH here */
	res = nis_list(sname, 
		       MASTER_ONLY+USE_DGRAM+NO_AUTHINFO+FOLLOW_LINKS,
		       NULL, NULL);
	
	status = res->status;
	switch( status ) {
	case NIS_NOTFOUND:
		break;
	case NIS_TRYAGAIN :
		fprintf(stderr, "%s: NIS+ server busy, try again later.\n",
			program_name);
		exit(1);
	case NIS_PERMISSION :
		fprintf(stderr, 
		"%s: insufficent permission to look at credentials table\n",
			program_name);
		exit(1);
	case NIS_SUCCESS:
	case NIS_S_SUCCESS:
		break;
	default:
		fprintf(stderr, 
			"%s: error looking at cred table, NIS+ error: %s\n",
			program_name, nis_sperrno(res->status));
		exit(1);
	}
	nis_freeresult(res);
	return( status );
}


static
int
modify_cred_obj( obj, domain )
	char* domain;
	nis_object *obj;
{
	int status = 0;
	int flags = 0;
	char sname[NIS_MAXNAMELEN+1];
	nis_result	*res;

	sprintf(sname, "%s.%s", CRED_TABLE, domain);
	res = nis_modify_entry(sname, obj, 0);
	switch (res->status) {
	case NIS_TRYAGAIN :
		fprintf(stderr, 
			"%s: NIS+ server busy, try again later.\n",
			program_name);
		exit(1);
	case NIS_PERMISSION :
		fprintf(stderr, 
			"%s: insufficent permission to update credentials.\n",
			program_name);
		exit(1);
	case NIS_SUCCESS :
		status = 1;
		break;
	default:
		fprintf(stderr, 
			"%s: error creating credential, NIS+ error: %s.\n",
			program_name, nis_sperrno(res->status));
		exit(1);
	}
	nis_freeresult(res);
	return( status );
}


static
int
add_cred_obj( obj, domain )
	char* domain;
	nis_object *obj;
{
	int status = 0;
	int flags = 0;
	char sname[NIS_MAXNAMELEN+1];
	nis_result	*res;

	/* Assume check for cred_exists performed already */

	sprintf(sname, "%s.%s", CRED_TABLE, domain);
	res = nis_add_entry(sname, obj, 0);
	switch (res->status) {
	case NIS_TRYAGAIN :
		fprintf(stderr, 
			"%s: NIS+ server busy, try again later.\n",
			program_name);
		exit(1);
	case NIS_PERMISSION :
		fprintf(stderr, 
			"%s: insufficent permission to update credentials.\n",
			program_name);
		exit(1);
	case NIS_SUCCESS :
		status = 1;
		break;
	default:
		fprintf(stderr, 
			"%s: error creating credential, NIS+ error: %s.\n",
			program_name, nis_sperrno(res->status));
		exit(1);
	}
	nis_freeresult(res);
	return( status );
}

nis_object *
init_entry()
{
	static nis_object	obj;
	static entry_col	cred_data[10];
	entry_obj		*eo;

	memset((char *)(&obj), 0, sizeof (obj));
	memset((char *)(cred_data), 0, sizeof (entry_col) * 10);
	
	obj.zo_name = "cred";
	obj.zo_group = "";
	obj.zo_ttl = 43200;
	obj.zo_data.zo_type = ENTRY_OBJ;
	eo = &(obj.EN_data);
	eo->en_type = "cred_tbl";
	eo->en_cols.en_cols_val = cred_data;
	eo->en_cols.en_cols_len = 5;
	cred_data[4].ec_flags |= EN_CRYPT;
	return (&obj);
}

int
nisplus_update( char* netname, char* public, char* secret )
{
	nis_object	*obj = init_entry();
	char* domain = nis_local_directory();
	int status, addition;
	char* nis_princ;

	nis_princ = default_principal( domain );

	if( sanity_checks( nis_princ, netname, domain ) == 0)
		return( 0 );

	addition = (cred_exists( nis_princ, "DES", domain ) == NIS_NOTFOUND);
	
	/* Now we have a key pair, build up the cred entry */
	ENTRY_VAL(obj, 0) = nis_princ;
	ENTRY_LEN(obj, 0) = strlen(nis_princ) + 1;

	ENTRY_VAL(obj, 1) = "DES";
	ENTRY_LEN(obj, 1) = 4;

	ENTRY_VAL(obj, 2) = netname;
	ENTRY_LEN(obj, 2) = strlen(netname) + 1;

	ENTRY_VAL(obj, 3) = public;
	ENTRY_LEN(obj, 3) = strlen(public) + 1;

	ENTRY_VAL(obj, 4) = secret;
	ENTRY_LEN(obj, 4) = strlen(secret) + 1;

	if( addition ) {
		obj->zo_owner = nis_princ;
		obj->zo_group = nis_local_group();
		obj->zo_domain = domain;
		/* owner: r, group: rmcd */
		obj->zo_access = ((NIS_READ_ACC<<16)|
				 (NIS_READ_ACC|NIS_MODIFY_ACC|NIS_CREATE_ACC|
				  NIS_DESTROY_ACC)<<8);
		status = add_cred_obj( obj, domain );
	} else {
		obj->EN_data.en_cols.en_cols_val[3].ec_flags |= EN_MODIFIED;
		obj->EN_data.en_cols.en_cols_val[4].ec_flags |= EN_MODIFIED;
		status = modify_cred_obj( obj, domain );
	}

	return( status == 1 ? 0 : -1);
}
