/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

#pragma ident "@(#)newkey.c	1.11	92/04/21 SMI"

/*
 * Copyright (C) 1986, Sun Microsystems, Inc.
 */

/*
 * Administrative tool to add a new user to the publickey database
 */
#include <stdio.h>
#include <rpc/rpc.h>
#include <rpc/key_prot.h>
#include <rpcsvc/ypclnt.h>
#include <sys/wait.h>
#include <netdb.h>
#include <pwd.h>
#include <string.h>
#include <sys/resource.h>
#include <netdir.h>

#define	MAXMAPNAMELEN 256
#define	YPOP_CHANGE 1			/* change, do not add */
#define	YPOP_INSERT 2			/* add, do not change */
#define	YPOP_DELETE 3			/* delete this entry */
#define	YPOP_STORE  4			/* add, or change */
#define	ERR_ACCESS	1
#define	ERR_MALLOC	2
#define	ERR_READ	3
#define	ERR_WRITE	4
#define	ERR_DBASE	5
#define	ERR_KEY		6

#define PK_FILES   1
#define PK_YP      2
#define PK_NISPLUS 3

extern char* get_password();
char program_name[256];

extern char *malloc();
extern int optind;
extern char *optarg;

static char *basename();
static char SHELL[] = "/bin/sh";
static char YPDBPATH[]="/var/yp";
static char PKMAP[] = "publickey.byname";
static char UPDATEFILE[] = "updaters";
static char PKFILE[] = "/etc/publickey";
static char *err_string();

main(argc, argv)
	int argc;
	char *argv[];
{
	char domain[MAXHOSTNAMELEN + 1];
	char name[MAXNETNAMELEN + 1];
	char public[HEXKEYBYTES + 1];
	char secret[HEXKEYBYTES + 1];
	char crypt1[HEXKEYBYTES + KEYCHECKSUMSIZE + 1];
	int status, pk_database;
	char *pass, *target_host = 0, *username = 0, *pk_service = 0;
	struct passwd *pw;
	NCONF_HANDLE *nc_handle;
	struct netconfig *nconf;
	struct nd_hostserv service;
	struct nd_addrlist *addrs;
	bool_t validhost;
	uid_t uid;
	char c;

	strcpy( program_name, argv[0] );
	while ((c = getopt(argc, argv, "s:u:h:")) != -1) {
		switch (c) {
		case 's':
			if( pk_service == NULL )
				pk_service = optarg;
			else
				usage(program_name);
			break;
		case 'u':
			if( username || target_host )
				usage(program_name);
			username = optarg;
			break;
		case 'h':
			if( username || target_host )
				usage(program_name);
			target_host = optarg;
			break;
		default:
			usage(program_name);
		}
	}

	if( optind < argc || (username == 0 && target_host == 0)) {
		usage(program_name);
	}

	if( (pk_database = get_pk_source( pk_service )) == 0 )
		usage(program_name);

	if (geteuid() != 0) {
		(void)fprintf(stderr, "Must be superuser to run %s\n", 
			      program_name);
		exit(1);
	}

	if ( username ) {
		pw = getpwnam(username);
		if (pw == NULL) {
			(void)fprintf(stderr, "%s: unknown user: %s\n", 
				      program_name, username);
			exit(1);
		}
		uid = pw->pw_uid;
		getdomainname( domain, MAXHOSTNAMELEN );
		getnetnameof(name, uid, domain);
	} else {
		/* -h hostname option */
		service.h_host = target_host;
		service.h_serv = NULL; 
		validhost = FALSE;
		/* verify if this is a valid hostname */
		nc_handle = setnetconfig();
		if (nc_handle == NULL) {
			/* fails to open netconfig file */
			fprintf(stderr, 
				"%s: failed in routine setnetconfig(0)\n",
				program_name);
			exit(2);
		}
		while (nconf = getnetconfig(nc_handle)) {
			/* check to see if hostname exists for this transport */
			if ( (netdir_getbyname(nconf, &service, &addrs) == 0) &&
			    (addrs->n_cnt != 0)) {
				/* atleast one valid address */
				validhost = TRUE;
				break;
			}
		}
		endnetconfig(nc_handle);
		if (!validhost) {
			(void)fprintf(stderr, "%s: unknown host: %s\n", 
				      program_name, target_host);
			exit(1);
		}
		(void)host2netname(name, target_host, (char *)NULL);
		uid = 0;
	}

	(void)printf("Adding new key for %s.\n", name);
	pass = get_password( uid, 0, target_host, 1 );
	genkeys(public, secret, pass);

	memcpy(crypt1, secret, HEXKEYBYTES);
	memcpy(crypt1 + HEXKEYBYTES, secret, KEYCHECKSUMSIZE);
	crypt1[HEXKEYBYTES + KEYCHECKSUMSIZE] = 0;
	xencrypt(crypt1, pass);

	if (status = setpublicmap(name, public, crypt1, pk_database)) {
		switch( pk_database ) {
		case PK_YP:
			(void)fprintf(stderr,
			      "%s: unable to update yp database (%u): %s\n",
				      program_name, status, 
				      yperr_string(status));
			break;
		case PK_FILES:
			(void)fprintf(stderr,
		"%s: unable to update publickey database (%u): %s\n",
			program_name, status, err_string(status));
			break;
		case PK_NISPLUS:
			(void)fprintf(stderr,
		      "%s: unable to update nisplus database\n", program_name);
			break;
		default:
			fprintf(stderr, 
				"%s: could not update unknown database: %d\n",
				program_name, pk_database);
		}
		exit(1);
	}
	(void)printf("Your new key has been successfully stored away.\n");
	exit(0);
	/* NOTREACHED */
}

/*
 * Set the entry in the public key file
 */
setpublicmap(name, public, secret, database)
	int database;
	char *name;
	char *public;
	char *secret;
{
	char pkent[1024];
	char* domain = NULL;
	char* master = NULL;
	char hostname[MAXHOSTNAMELEN+1];

	(void)sprintf(pkent, "%s:%s", public, secret);
	switch( database ) {
	case PK_YP:
		/* check that we're on the master server */
		(void)yp_get_default_domain(&domain);
		if (yp_master(domain, PKMAP, &master) != 0) {
			(void)fprintf(stderr,
			      "%s: cannot find master of publickey database\n",
				      program_name);
			exit(1);
		}
		if( gethostname( hostname, MAXHOSTNAMELEN ) < 0 ) {
			fprintf(stderr, "%s: cannot find my own host name\n",
				program_name);
			exit(1);
		}
		if( strcmp(master,hostname) != 0 ) {
			fprintf( stderr,
				"%s: can only be used on master machine '%s'\n",
				program_name, master);
			exit(1);
		}

		if (chdir(YPDBPATH) < 0) {
			(void)fprintf(stderr, "%s: cannot chdir to %s",
				      program_name, YPDBPATH);
		}
		printf("Please wait for the database to get updated...\n");
		return (mapupdate(name, PKMAP, YPOP_STORE,
				  strlen(name), name, strlen(pkent), pkent));
	case PK_FILES:
		return (localupdate(name, PKFILE, YPOP_STORE,
				    strlen(name), name, strlen(pkent), pkent));
	case PK_NISPLUS:
		return( nisplus_update( name, public, secret ) );
	default:
		fprintf( stderr, 
			"%s: don't know which database to update (%d)\n",
			program_name, database);
	}
}


/*
 * This returns a pointer to an error message string appropriate
 * to an input error code.  An input value of zero will return
 * a success message.
 */
static char *
err_string(code)
	int code;
{
	char *pmesg;

	switch (code) {
	case 0:
		pmesg = "update operation succeeded";
		break;
	case ERR_KEY:
		pmesg = "no such key in file";
		break;
	case ERR_READ:
		pmesg = "cannot read the database";
		break;
	case ERR_WRITE:
		pmesg = "cannot write to the database";
		break;
	case ERR_DBASE:
		pmesg = "cannot update database";
		break;
	case ERR_ACCESS:
		pmesg = "permission denied";
		break;
	case ERR_MALLOC:
		pmesg = "malloc failed";
		break;
	default:
		pmesg = "unknown error";
		break;
	}
	return (pmesg);
}

usage( char* prog )
{
	(void)fprintf(stderr, 
		      "usage: %s [-u username] [-s nisplus | nis | files]\n", 
		      prog );
	(void)fprintf(stderr, 
		      "usage: %s [-h hostname] [-s nisplus | nis | files]\n", 
		      prog );
	exit(1);
}
